typedef struct 
{
   int x;
}
Message_Box_Type;

typedef struct 
{
   int curr;			       /* currently selected. */
   char *buf;		       /* \0 separated */
   char **item_ptrs;		       /* pointers into above field */
   int item_ptr_size;		       /* number of such pointers available*/
   int nitems;			       /* number used */
   int top_item;		       /* item number of top visible */
}
List_Box_Type;

typedef struct 
{
#define MAX_EDITBOX_SIZE 256
   char buf[MAX_EDITBOX_SIZE];
   int point;
}
Edit_Box_Type;

typedef struct 
{
   char *text;
   int lb_id;
}
Popup_ListBox_Type;

typedef struct SLWdg_Type
{
   int id;							/* id used by client */
   unsigned int flags;
#define HELP_SHOWING 1
#define NO_FOCUS 2
#define LIST_BOX_ACTIVE 4
#define WIDGET_NOT_MAPPED 8
#define PERFORMING_CALLBACK 16
   char *label;
   char *help;
   char *summary;		       /* short summary of what object does */
   int x, y;
   int width, height;
   
   void (*draw) (struct SLWdg_Type *);
   int (*focus_proc) (struct SLWdg_Type *, int);
   /* focus_proc returns zero if the widget handled the event.  Otherwise, 
    * it returns non-zero and the event is passed to the parent for 
    * processing.
    */
   int (*resize_widget) (struct SLWdg_Type *, int, int, int, int);
   
   int (*callback) (int, int, char *);
   int callback_int;
   char *callback_str;
   int (*set_text) (struct SLWdg_Type *, char *);
   char *(*get_text) (struct SLWdg_Type *);
   
   struct SLWdg_Type *next;
   struct SLWdg_Type *prev;
   struct SLWdg_Type *child;
   struct SLWdg_Type *current;	      /* points to CURRENT child */
   struct SLWdg_Type *parent;
   struct SLWdg_Type *last;	       /* points to last active before popup */
   
   unsigned int type;		       /* EDIT_BOX, etc... */
   union
     {
	Edit_Box_Type *editbox;
	List_Box_Type *listbox;
	Message_Box_Type *msgbox;
	Popup_ListBox_Type *popbox;
     }
   widget;
   unsigned int num_references;
   int do_free;
}
SLWdg_Type;

extern SLWdg_Type *SLWdg_Root_Widget;

extern void SLwdg_draw_base_widget (SLWdg_Type *);
extern void SLwdg_update_display (void);
extern void SLwdg_take_input (int);
extern void SLwdg_delete_widget (SLWdg_Type *);

extern SLWdg_Type *SLwdg_validate_id (int, unsigned int);
extern int SLwdg_focus_next (SLWdg_Type *);
extern SLWdg_Type *SLwdg_find_widget_at_xy (SLWdg_Type *, int, int);
extern int SLwdg_read_mouse_keys (void);
extern int SLWdg_Mouse_X;
extern int SLWdg_Mouse_Y;

extern int SLwdg_init_listbox (SLWdg_Type *);
extern int SLwdg_init_editbox (SLWdg_Type *);
extern int SLwdg_init_button (SLWdg_Type *);
/* extern int SLwdg_init_msgbox (SLWdg_Type *); */
extern int SLwdg_init_combobox (SLWdg_Type *);
extern int SLwdg_init_combobox (SLWdg_Type *);
extern int SLwdg_init_popupbox (SLWdg_Type *);
extern int SLwdg_perform_callback (SLWdg_Type *);

extern void SLwdg_suspend (int);


/* These are the keysyms or messages */
#define KEY_CTRL_G      0x007
#define KEY_CTRL_R	0x012
#define KEY_CTRL_Z	0x01A
#define KEY_RETURN      0x00D
#define KEY_TAB		0x009
#define KEY_BACKSPACE	0x008
#define KEY_DELETE	0x07F
#define KEY_DOWN_ARROW  0x101
#define KEY_UP_ARROW    0x102
#define KEY_LEFT_ARROW  0x103
#define KEY_RIGHT_ARROW 0x104
#define KEY_MOUSE0	0x201
#define KEY_MOUSE1	0x202
#define KEY_MOUSE2	0x203
#define SLW_LOSE_FOCUS	0x301
#define SLW_GAIN_FOCUS	0x302
#define SLW_DESTROY	0x303
#define SERVER_ESCAPE	0x7FFF
