#include <stdio.h>
#include <string.h>

#include "slang.h"
#include "slwdg.h"
#include "_slwdg.h"


void SLwdg_draw_base_widget (SLWdg_Type *ip)
{
   SLWdg_Type *parent = ip->parent;
   
   if (ip == ip->parent) return;
   
   if (ip == ip->parent->current)
     SLsmg_set_color (ACTIVE_BOX_COLOR);
   else SLsmg_set_color (BOX_COLOR);
   SLsmg_draw_box (ip->y, ip->x, ip->height, ip->width);
   
   if (ip == parent->current) SLsmg_set_color (ACTIVE_LABEL_COLOR);
   else SLsmg_set_color (LABEL_COLOR);
   
   if (ip->label != NULL)
     {
	SLsmg_gotorc (ip->y, ip->x + 1);
	SLsmg_printf ("[%s]", ip->label);
     }
   
   SLsmg_set_color (0);
   SLsmg_fill_region (ip->y + 1, ip->x + 1, ip->height - 2, ip->width - 2, ' ');
}

static void draw_combo_box (SLWdg_Type *ip)
{
   SLWdg_Type *curr_ip, *this_ip;

   this_ip = ip;
   
   if (ip->id == 0)
     {
	/* root widget */
	SLsmg_set_color (0);
	SLsmg_gotorc (0, 0);
	SLsmg_erase_eos ();
     }
   else SLwdg_draw_base_widget (ip);
   
   ip = curr_ip = ip->current;
   
   if (ip != NULL) 
     {
	ip = ip->next;
	do 
	  {
	     if ((ip->flags & WIDGET_NOT_MAPPED) == 0)
	       {
		  if (ip == curr_ip)
		    {
		       SLsmg_gotorc (SLtt_Screen_Rows - 1, 0);
		       SLsmg_set_color (SUMMARY_COLOR);
		       if (ip->summary != NULL) SLsmg_write_string (ip->summary);
		       SLsmg_erase_eol ();
		       SLsmg_set_color (0);
		    }
		  
		  /* The recursive situation avoided here can occur if
		   * this ip is the root.
		   */
		  if ((ip->draw != draw_combo_box)
		      || (ip != this_ip))
		    (*ip->draw)(ip);
		  
		  /*
		  if (ip->flags & HELP_SHOWING)
		    {
		       draw_help (ip);
		    } */
	       }
	     
	     ip = ip->next;
	  }
	while (ip != curr_ip->next);
     }
}


static int resize_combobox (SLWdg_Type *ip, int x, int y, int dx, int dy)
{
   SLWdg_Type *parent, *child;
   int delta_x, delta_y;
   /* Root is always at 0, 0 */
   if (ip->id != 0)
     {
	int cx, cy;
	parent = ip->parent;
	cx = ip->x;
	cy = ip->y;
	delta_x = x - (cx - parent->x);
	delta_y = y - (cy - parent->y);
	child = ip->child;
	do 
	  {
	     if (-1 == (*child->resize_widget) (child, child->x - cx + delta_x,
						child->y - cy + delta_y,
						child->width, child->height))
	       return -1;
	     child = child->next;
	  }
	while (child != ip->child);
	ip->x += delta_x;
	ip->y += delta_y;
     }
   ip->width = dx;
   ip->height = dy;
   return 0;
}

static void find_widget_up (SLWdg_Type *this_ip)
{
   SLWdg_Type *ip, *best_ip;
   int x, y, x1;
   int min_x, max_x, max_y;
   int best_y = -1;
   
   
   min_x = this_ip->x;
   max_x = min_x + this_ip->width;
   max_y = this_ip->y + this_ip->height / 2;
   
   best_ip = ip = this_ip->prev;
   
   if ((best_ip->flags & NO_FOCUS) || (best_ip->flags & WIDGET_NOT_MAPPED)) 
     best_ip = this_ip;
   
   while (ip != this_ip)
     {
	if (((ip->flags & NO_FOCUS) == 0)
	    && ((ip->flags & WIDGET_NOT_MAPPED) == 0)
	    && ((y = ip->y) < max_y)
	    && (y > best_y))
	  {
	     x = ip->x; x1 = x + ip->width;
	     if (((x >= min_x) && (x < max_x))
		 || ((x1 >= min_x) && (x1 < max_x))
		 || ((x <= min_x) && (x1 >= max_x)))
	       {
		  best_ip = ip;
		  best_y = y;
	       }
	  }
	ip = ip->prev;
     }
   
   if (best_ip != this_ip->parent->current)
     {
	(*this_ip->focus_proc) (this_ip->parent->current, SLW_LOSE_FOCUS);
	(*this_ip->focus_proc) (best_ip, SLW_GAIN_FOCUS);
	this_ip->parent->current = best_ip;
     }
   else SLwdg_focus_next (this_ip);
}

static void find_widget_left (SLWdg_Type *this_ip)
{
   SLWdg_Type *ip, *best_ip;
   int x, y, y1;
   int min_y, max_y, max_x;
   int best_x = -1;
   
   
   min_y = this_ip->y;
   max_y = min_y + this_ip->height;
   max_x = this_ip->x;
   
   best_ip = ip = this_ip->prev;
   if ((best_ip->flags & NO_FOCUS) || (best_ip->flags & WIDGET_NOT_MAPPED)) 
     best_ip = this_ip;
   
   while (ip != this_ip)
     {
	if (((ip->flags & NO_FOCUS) == 0)
	    && ((ip->flags & WIDGET_NOT_MAPPED) == 0)
	    && ((x = ip->x) < max_x)
	    && (x > best_x))
	  {
	     y = ip->y; y1 = y + ip->height;
	     if (((y >= min_y) && (y < max_y))
		 || ((y1 > min_y) && (y1 <= max_y))
		 || ((y <= min_y) && (y1 >= max_y)))
	       {
		  best_ip = ip;
		  best_x = x;
	       }
	  }
	ip = ip->prev;
     }
   
   if (best_ip != this_ip->parent->current)
     {
	(*this_ip->focus_proc) (this_ip->parent->current, SLW_LOSE_FOCUS);
	(*this_ip->focus_proc) (best_ip, SLW_GAIN_FOCUS);
	this_ip->parent->current = best_ip;
     }
   else SLwdg_focus_next (this_ip);
}

static void find_widget_right (SLWdg_Type *this_ip)
{
   SLWdg_Type *ip, *best_ip;
   int x, y, y1;
   int min_y, max_y, min_x;
   int best_x = 0x7FFF;
   
   
   min_y = this_ip->y;
   max_y = min_y + this_ip->height;
   min_x = this_ip->x + this_ip->width / 2;
   
   best_ip = ip = this_ip->next;
   if ((best_ip->flags & NO_FOCUS) || (best_ip->flags & WIDGET_NOT_MAPPED)) 
     best_ip = this_ip;
   
   while (ip != this_ip)
     {
	if (((ip->flags & NO_FOCUS) == 0)
	    && ((ip->flags & WIDGET_NOT_MAPPED) == 0)
	    && ((x = ip->x) > min_x)
	    && (x < best_x))
	  {
	     y = ip->y; y1 = y + ip->height;
	     if (((y >= min_y) && (y < max_y))
		 || ((y1 > min_y) && (y1 <= max_y))
		 || ((y <= min_y) && (y1 >= max_y)))
	       {
		  best_ip = ip;
		  best_x = x;
	       }
	  }
	ip = ip->next;
     }
   if (best_ip != this_ip->parent->current)
     {
	(*this_ip->focus_proc) (this_ip->parent->current, SLW_LOSE_FOCUS);
	(*this_ip->focus_proc) (best_ip, SLW_GAIN_FOCUS);
	this_ip->parent->current = best_ip;
     }
   else SLwdg_focus_next (this_ip);
}

static void find_widget_down (SLWdg_Type *this_ip)
{
   SLWdg_Type *ip, *best_ip;
   int x, y, x1;
   int min_x, max_x, min_y;
   int best_y = 0x7FFF;
   
   
   min_x = this_ip->x;
   max_x = min_x + this_ip->width;
   min_y = this_ip->y + this_ip->height / 2;
   
   best_ip = ip = this_ip->next;
   
   if ((best_ip->flags & NO_FOCUS) || (best_ip->flags & WIDGET_NOT_MAPPED)) 
     best_ip = this_ip;
   
   while (ip != this_ip)
     {
	if (((ip->flags & NO_FOCUS) == 0)
	    && ((ip->flags & WIDGET_NOT_MAPPED) == 0)
	    && ((y = ip->y) > min_y)
	    && (y < best_y))
	  {
	     x = ip->x; x1 = x + ip->width;
	     if (((x >= min_x) && (x < max_x))
		 || ((x1 >= min_x) && (x1 < max_x))
		 || ((x <= min_x) && (x1 > max_x)))
	       {
		  best_ip = ip;
		  best_y = y;
	       }
	  }
	ip = ip->next;
     }

   if (best_ip != this_ip->parent->current)
     {
	(*this_ip->focus_proc) (this_ip->parent->current, SLW_LOSE_FOCUS);
	(*this_ip->focus_proc) (best_ip, SLW_GAIN_FOCUS);
	this_ip->parent->current = best_ip;
     }
   else SLwdg_focus_next (this_ip);
}


static int Server_Escape_Mode = 0;
static int Server_Escape_Mod_Width = 1;
static void move_widget_dir (SLWdg_Type *ip, int dir)
{
   int x, y, dx, dy;
   
   (void) SLwdg_get_size (ip->id, &x, &y, &dx, &dy);
   switch (dir * Server_Escape_Mod_Width)
     {
      case KEY_RIGHT_ARROW:
	if (x < SLtt_Screen_Cols - 2) x++;
	break;
      case KEY_LEFT_ARROW:
	if (x > 0) x--;
	break;
	
      case KEY_UP_ARROW:
	if (y > 0) y--;
	break;
      case KEY_DOWN_ARROW:
	if (y < SLtt_Screen_Rows - 2) y++;
	break;
      case -KEY_RIGHT_ARROW:
	dx++;
	break;
      case -KEY_LEFT_ARROW:
	if (dx > 1) dx--;
	break;
	
      case -KEY_UP_ARROW:
	if (dy > 1) dy--;
	break;
      case -KEY_DOWN_ARROW:
	dy++;
	break;
     }
   (void) SLwdg_set_size (ip->id, x, y, dx, dy);
}
   
static int combobox_focus_proc (SLWdg_Type *this_ip, int keysym)
{
   SLWdg_Type *curr, *ip;
   
   
   if ((curr = this_ip->current) == NULL) 
     {
	SLwdg_exit_error ("COMBOBOX[%d] is empty.  Cannot take focus.", 
			  this_ip->id);
     }
   
   if (Server_Escape_Mode && (curr->focus_proc != combobox_focus_proc))
     {
	switch (keysym)
	  {
	   case KEY_RIGHT_ARROW:
	   case KEY_LEFT_ARROW:
	   case KEY_UP_ARROW:
	   case KEY_DOWN_ARROW:
	     if (Server_Escape_Mode > 0)
	       {
		  move_widget_dir (curr, keysym);
	       }
	     else 
	       move_widget_dir (this_ip, keysym);
	     break;
		  
	   case ' ':
	     Server_Escape_Mode = -Server_Escape_Mode;
	     break;
	   case 'w':
	     Server_Escape_Mod_Width = -Server_Escape_Mod_Width;
	     break;
	   default:
	     Server_Escape_Mode = 0;
	  }
	return 0;
     }

   this_ip->num_references++;
   if (0 == ((*curr->focus_proc) (curr, keysym))) 
     {
	this_ip->num_references--;
	if (this_ip->do_free) SLwdg_delete_widget (this_ip);
	return 0;
     }
   
   switch (keysym)
     {
      case SERVER_ESCAPE:
	Server_Escape_Mode = !Server_Escape_Mode;
	return 0;
	
      case KEY_TAB:
      case KEY_RETURN:
	SLwdg_focus_next (curr);
	break;
	
      case KEY_RIGHT_ARROW:
	find_widget_right (curr);
	break;
	
      case KEY_LEFT_ARROW:
	find_widget_left (curr);
	break;
	
      case KEY_DOWN_ARROW:
	find_widget_down (curr);
	break;
	
      case KEY_UP_ARROW:
	find_widget_up (curr);
	break;
	
	
      case KEY_MOUSE0:
      case KEY_MOUSE1:
      case KEY_MOUSE2:
	ip = SLwdg_find_widget_at_xy (curr, SLWdg_Mouse_X, SLWdg_Mouse_Y);
	if (ip != NULL) 
	  {
	     if (ip != curr)
	       {
		  (*curr->focus_proc) (curr, SLW_LOSE_FOCUS);
		  this_ip->current = ip;
		  (*curr->focus_proc) (curr, SLW_GAIN_FOCUS);
		  (void) (*ip->focus_proc) (ip, keysym);
	       }
	  }
	break;
	     
      case KEY_CTRL_Z:
	SLwdg_suspend (0);
	break;
	
      case SLW_DESTROY:
	while (this_ip->child != NULL) SLwdg_destroy_widget (this_ip->child->id);
	this_ip->current = NULL;
	break;
	
      default:
	this_ip->num_references--;
	if (this_ip->do_free) SLwdg_delete_widget (this_ip);
	return -1;
     }
   this_ip->num_references--;
   if (this_ip->do_free) SLwdg_delete_widget (this_ip);
   return 0;
}

int SLwdg_init_combobox (SLWdg_Type *ip)
{
   ip->draw = draw_combo_box;
   ip->focus_proc = combobox_focus_proc;
   ip->resize_widget = resize_combobox;
   return 0;
}
