#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <slang.h>

#include "slwdg.h"
#include "_slwdg.h"

static void draw_edit_box (SLWdg_Type *ip)
{
   SLwdg_draw_base_widget (ip);
   SLsmg_gotorc (ip->y + 1, ip->x + 1);
   
   if (ip == ip->parent->current)
     SLsmg_set_color (ACTIVE_FIELD_COLOR);
   else
     SLsmg_set_color (FIELD_COLOR);
   
   SLsmg_write_nstring (ip->widget.editbox->buf, ip->width - 2);
   SLsmg_gotorc (ip->y + 1, ip->x + 1);
   
   SLsmg_set_color (0);
}

static SLWdg_Type *Current_Edit_Widget;
static void rline_update (unsigned char *buf, int len, int col)
{
   SLWdg_Type *ip = Current_Edit_Widget;
   SLsmg_gotorc (ip->y + 1, ip->x + 1);
   SLsmg_set_color (FIELD_COLOR);
   SLsmg_write_nchars ((char *) buf, len);
   SLsmg_gotorc (ip->y + 1, ip->x + col + 1);
   SLsmg_refresh ();
}

static int tab_next_field (void)
{
   SLang_Rline_Quit = 1;
   return 0;
}

SLang_RLine_Info_Type *SLwdg_Keymap_RLI;

static int edit_suspend (void)
{
   SLwdg_suspend (0);
   SLrline_redraw (SLwdg_Keymap_RLI);
   return 0;
}

static SLWdg_Type *Next_Focus;
static int Mouse_Key;

static int edit_handle_mouse (void)
{
   SLWdg_Type *ip;
   
   Mouse_Key = SLwdg_read_mouse_keys ();
   ip = SLwdg_find_widget_at_xy (Current_Edit_Widget, 
				    SLWdg_Mouse_X, SLWdg_Mouse_Y);
   if (ip == NULL) return 0;
   if (ip != ip->parent->current)
     {
	Next_Focus = ip;
	SLang_Rline_Quit = 1;
     }
   else Mouse_Key = 0;
   return 0;
}

   
static SLang_RLine_Info_Type *init_readline (void)
{
   unsigned char *buf = NULL;
   SLang_RLine_Info_Type *rli;
   
   if ((NULL == (rli = (SLang_RLine_Info_Type *) SLMALLOC (sizeof(SLang_RLine_Info_Type))))
       || (NULL == (buf = (unsigned char *) SLMALLOC (MAX_EDITBOX_SIZE))))
     {
	fprintf(stderr, "malloc error.\n");
	exit(-1);
     }
   
   memset ((char *) rli, 0, sizeof (SLang_RLine_Info_Type));
   rli->buf = buf;
   rli->buf_len = MAX_EDITBOX_SIZE - 1;
   rli->tab = 8;
   rli->dhscroll = 1;
   rli->getkey = SLang_getkey;
   rli->tt_goto_column = NULL;
   rli->update_hook = rline_update;
   
   if ((SLang_init_readline (rli) < 0)
       || SLkm_define_key ("\t", (FVOID_STAR) tab_next_field, rli->keymap)
       || SLkm_define_key ("^Z", (FVOID_STAR) edit_suspend, rli->keymap)
       || SLkm_define_key ("\033[M", (FVOID_STAR) edit_handle_mouse, rli->keymap)
       )
     {
	fprintf(stderr, "Unable to initialize readline library.\n");
	exit (-1);
     }

   return rli;
}

static int read_input_from_field (SLWdg_Type *ip, int set_point, int clear)
{
   int i;
   
   if (SLwdg_Keymap_RLI == NULL) 
     {
	SLwdg_Keymap_RLI = init_readline ();
     }

   try_again:
   
   Mouse_Key = 0;
   Current_Edit_Widget = ip;
   
   SLwdg_Keymap_RLI->edit_width = ip->width - 2;
   SLwdg_Keymap_RLI->prompt = NULL;
   if (clear == 0)
     {
	strcpy ((char *) SLwdg_Keymap_RLI->buf, ip->widget.editbox->buf);
	if (set_point)
	  SLwdg_Keymap_RLI->point = ip->widget.editbox->point;
	else SLwdg_Keymap_RLI->point = 0;
     }
   else
     {
	*SLwdg_Keymap_RLI->buf = 0;
	SLwdg_Keymap_RLI->point = 0;
     }
   
   i = SLang_read_line (SLwdg_Keymap_RLI);
   
   if ((i >= 0) && !SLang_Error && !SLKeyBoard_Quit)
     {
	ip->widget.editbox->point = SLwdg_Keymap_RLI->point;
	strncpy (ip->widget.editbox->buf, (char *) SLwdg_Keymap_RLI->buf, MAX_EDITBOX_SIZE - 1);
	ip->widget.editbox->buf[MAX_EDITBOX_SIZE - 1] = 0;
	
	if (ip->callback != NULL)
	  {
	     if (SLwdg_perform_callback (ip) == -1)
	       {
		  clear = 0;
		  Next_Focus = NULL;
		  set_point = 1;
		  goto try_again;
	       }
	  }
     }
   if (SLKeyBoard_Quit) i = -1;
   SLang_Error = SLKeyBoard_Quit = 0;
   return i;
}

static int editbox_focus_proc (SLWdg_Type *ip, int keysym)
{
   if (keysym > 255) return -1;
   
   Next_Focus = NULL;
   switch (keysym)
     {
      case KEY_CTRL_Z:
      case KEY_CTRL_R:
      case KEY_TAB:
	return -1;
	
      case KEY_RETURN:
	read_input_from_field (ip, 1, 0);
	if (Next_Focus == NULL) SLwdg_focus_next (ip);
	else SLwdg_set_focus (Next_Focus->id);
	break;
	
      case SLW_DESTROY:
	if (ip->widget.editbox != NULL)
	  SLFREE (ip->widget.editbox);
	ip->widget.editbox = NULL;
	break;
	
      default:
	SLang_ungetkey (keysym);
	read_input_from_field (ip, 0, (keysym >= ' '));
	if (Next_Focus == NULL) SLwdg_focus_next (ip);
	else SLwdg_set_focus (Next_Focus->id);
     }
   return 0;
}

static int set_edit_text (SLWdg_Type *ip, char *s)
{
   char *buf = ip->widget.editbox->buf;
   strncpy (buf, s, MAX_EDITBOX_SIZE - 1);
   buf[MAX_EDITBOX_SIZE - 1] = 0;
   return 0;
}

static char *get_edit_text (SLWdg_Type *ip)
{
   return ip->widget.editbox->buf;
}


int SLwdg_init_editbox (SLWdg_Type *ip)
{
   Edit_Box_Type *eb;
   
   if (NULL == (eb = (Edit_Box_Type *) SLMALLOC (sizeof (Edit_Box_Type))))
     {
	return -1;
     }
   memset ((char *) eb, 0, sizeof (Edit_Box_Type));
   ip->widget.editbox = eb;
   
   ip->draw = draw_edit_box;
   ip->focus_proc = editbox_focus_proc;
   if (ip->height < 3) ip->height = 3;
   ip->set_text = set_edit_text;
   ip->get_text = get_edit_text;
   return 0;
}
