#include <stdio.h>
#include <string.h>

#include <stdlib.h>

#include <slang.h>

#include "slwdg.h"
#include "_slwdg.h"


static void draw_list_box (SLWdg_Type *ip)
{
   int r, c, dr, dc;
   char **item_ptrs;
   int curr, i, save_row, nitems, imax;
   List_Box_Type *lb;
   
   lb = ip->widget.listbox;
   item_ptrs = lb->item_ptrs;
   curr = lb->curr;
   nitems = lb->nitems;
   
   r = ip->y;
   c = ip->x;
   SLsmg_gotorc (r, c);
   SLsmg_set_color (0);
   
   dc = ip->width - 2;
   dr = nitems;
   
   if (dr > ip->height) dr = ip->height;

   if (r + dr + 2 >=  SLtt_Screen_Rows)
     {
	r = SLtt_Screen_Rows - dr - 3;
     }
   if (c + dc + 2 >= SLtt_Screen_Cols)
     {
	c = SLtt_Screen_Cols - dc - 3;
     }
   
   if (ip == ip->parent->current)
     SLsmg_set_color (ACTIVE_BOX_COLOR);
   else SLsmg_set_color (BOX_COLOR);
   SLsmg_draw_box (r, c, dr + 2, dc + 2);
   SLsmg_set_color (FIELD_COLOR);
   
   c++;
   save_row = -1;
   if ((curr < lb->top_item) || (lb->top_item + dr <= curr))
     {
	lb->top_item = curr - dr/2;
	if (lb->top_item < 0) lb->top_item = 0;
     }
   imax = lb->top_item + dr;
   
   for (i = lb->top_item; i < imax; i++)
     {
	r++;
	if (curr == i)
	  {
	     save_row = r;
	     continue;
	  }
	SLsmg_gotorc (r, c);
	if (i >= nitems)
	  SLsmg_write_nstring (NULL, dc);
	else
	  SLsmg_write_nstring (item_ptrs[i], dc);
     }
   
   if (save_row != -1)
     {
	SLsmg_set_color (SELECTION_COLOR);
	SLsmg_gotorc (save_row, c);
	SLsmg_write_nstring (item_ptrs[curr], dc);
	SLsmg_gotorc (save_row, c);
     }
   SLsmg_set_color (0);
}

static int set_listbox_text (SLWdg_Type *ip, char *text)
{
   List_Box_Type *lb;
   int len;
   int nitems, max_items, max_len;
   char *p, ch, *b, *buf, *blast;
   char **item_ptrs;
   
   lb = ip->widget.listbox;
   buf = lb->buf;
   
   if (buf != NULL)
     {
	SLFREE (buf);
	lb->buf = NULL;
	lb->nitems = 0;
	lb->curr = 0;
     }
   len = strlen (text);
   
   if (NULL == (buf = (char *) SLMALLOC (len + 1)))
     {
	return -1;
     }
   
   max_items = lb->item_ptr_size;
   if (max_items == 0)
     {
	max_items = 256;
	item_ptrs = (char **) SLMALLOC (max_items * sizeof (char **));
	if (item_ptrs == NULL)
	  {
	     SLFREE (buf);
	     return -1;
	  }
     }
   
   nitems = 0;
   p = text;
   max_len = 0;
   b = buf;
   
   while (1)
     {
	blast = b;
	item_ptrs[nitems++] = blast;
	while ((0 != (ch = *p++)) && (ch != '\n'))
	  {
	     *b++ = ch;
	  }
	if ((int) (b - blast) > max_len) max_len = (int) (b - blast);
	*b++ = 0;
	if (ch == 0) break;
	     
	if (nitems == max_items)
	  {
	     char **new_item_ptrs;
	     max_items = max_items += 256;
	     new_item_ptrs = (char **) SLREALLOC (item_ptrs, max_items * sizeof (char **));
	     if (new_item_ptrs == NULL)
	       {
		  SLFREE (buf);
		  SLFREE (item_ptrs);
		  return -1;
	       }
	     item_ptrs = new_item_ptrs;
	  }
     }
   lb->buf = buf;
   lb->curr = 0;
   lb->item_ptrs = item_ptrs;
   lb->item_ptr_size = max_items;
   lb->nitems = nitems;
   ip->width = max_len + 2;	       /* includes border */
   return 0;
}

int SLwdg_set_listbox_items (int id, char *list)
{
   SLWdg_Type *ip;
   if (NULL == (ip = SLwdg_validate_id (id, LIST_BOX)))
     {
	return -1;
     }
   return set_listbox_text (ip, list);
}


static int lb_select (SLWdg_Type *ip)
{
   SLwdg_perform_callback (ip);
   return 0;
}


static int lb_mouse_select (SLWdg_Type *ip)
{
   List_Box_Type *lb = ip->widget.listbox;
   int x, y, xmax, ymax, dy;
   
   x = ip->x + 1;
   y = ip->y + 1;
   
   dy = lb->nitems;
   if (dy > ip->height) dy = ip->height;
   
   xmax = x + ip->width - 2;
   ymax = y + dy;
   
   if ((SLWdg_Mouse_Y < y) || (SLWdg_Mouse_Y >= ymax)
       || (SLWdg_Mouse_X < x) || (SLWdg_Mouse_X >= xmax))
     return -1;
   
   lb->curr = lb->top_item + (SLWdg_Mouse_Y - y);
   return 0;
}

       
static int lb_down (SLWdg_Type *ip)
{
   int curr;
   List_Box_Type *lb = ip->widget.listbox;
   curr = lb->curr;
   curr++;
   if (curr >= lb->nitems) curr = 0;
   lb->curr = curr;
   return 0;
}

static int lb_up (SLWdg_Type *ip)
{
   int curr;
   List_Box_Type *lb = ip->widget.listbox;
   curr = lb->curr;
   curr--;
   if (curr < 0) curr = lb->nitems - 1;
   lb->curr = curr;
   return 0;
}

static int listbox_focus_proc (SLWdg_Type *ip, int keysym)
{
   List_Box_Type *lb;
   switch (keysym)
     {
      case KEY_UP_ARROW:
	lb_up (ip);
	break;
	
      case KEY_DOWN_ARROW:
	lb_down (ip);
	break;
	
      case SLW_LOSE_FOCUS:
      case KEY_RETURN:
	lb_select (ip);
	break;
	
      case KEY_MOUSE0:
      case KEY_MOUSE1:
      case KEY_MOUSE2:
	return lb_mouse_select (ip);
	
      case KEY_CTRL_Z:
      case KEY_CTRL_R:
	return -1;
	
      case SLW_DESTROY:
	lb = ip->widget.listbox;
	if (lb != NULL)
	  {
	     if (lb->buf != NULL) SLFREE (lb->buf);
	     if (lb->item_ptrs != NULL) SLFREE (lb->item_ptrs);
	     SLFREE (lb);
	     ip->widget.listbox = NULL;
	  }
	break;
	
      default:
	return -1;
     }
   return 0;
}

static char *get_text (SLWdg_Type *ip)
{
   List_Box_Type *lb = ip->widget.listbox;
   
   if ((lb == NULL) || (lb->nitems == 0)) return NULL;
   return lb->item_ptrs[lb->curr];
}

int SLwdg_lb_set_selection (int id, char *s)
{
   List_Box_Type *lb;
   SLWdg_Type *ip;
   int n, i;
   
   if (NULL == (ip = SLwdg_validate_id (id, LIST_BOX)))
     {
	return -1;
     }
   lb = ip->widget.listbox;
   if (lb == NULL) return -1;
   
   n = lb->nitems;
   for (i = 0; i < n; i++)
     { 
	if (!strcmp (s, lb->item_ptrs[i]))
	  {
	     lb->curr = i;
	     return 0;
	  }
     }
   return -1;
}



int SLwdg_init_listbox (SLWdg_Type *ip)
{
   List_Box_Type *lb;
      
   if (NULL == (lb = (List_Box_Type *) SLMALLOC (sizeof (List_Box_Type))))
     {
	return -1;
     }
   memset ((char *) lb, 0, sizeof (List_Box_Type));
   ip->widget.listbox = lb;
   ip->draw = draw_list_box;
   ip->focus_proc = listbox_focus_proc;
   ip->get_text = get_text;
   if (ip->height < 1) ip->height = 1;
   if (ip->summary == NULL) ip->summary = "Press Return to make selection.";
   return 0;
}


