#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <slang.h>
#include "slwdg.h"
#include "_slwdg.h"


static void draw_popup_box (SLWdg_Type *ip)
{
   int width = ip->width - 2;
   
   SLwdg_draw_base_widget (ip);
   
   SLsmg_gotorc (ip->y + ip->height - 1, ip->x + width - 2);
   SLsmg_write_string ("[+]");
   
   SLsmg_gotorc (ip->y + 1, ip->x + 1);
   if (SLtt_Use_Ansi_Colors) 
     {
	if (ip == ip->parent->current)
	  SLsmg_set_color (ACTIVE_FIELD_COLOR);
	else
	  SLsmg_set_color (FIELD_COLOR);
     }
   else SLsmg_set_color (0);

   SLsmg_write_nstring (ip->widget.popbox->text, width);
   SLsmg_gotorc (ip->y + 1, ip->x + 1);
   SLsmg_set_color (0);
}

static int set_popbox_text (SLWdg_Type *ip, char *text)
{
   Popup_ListBox_Type *pb;
   int len;
   
   pb = ip->widget.popbox;
   
   if (pb->text != NULL)
     {
	SLFREE (pb->text);
     }
   len = strlen (text);
   
   if (NULL == (pb->text = (char *) SLMALLOC (len + 1)))
     {
	return -1;
     }
   strcpy (pb->text, text);
   return 0;
}


static int pb_select (SLWdg_Type *ip, int state)
{
   (void) SLwdg_perform_callback (ip);
   return 0;
}


static int pb_mouse_select (SLWdg_Type *ip)
{
   int x, y;
   int on_icon;
   Popup_ListBox_Type *pb;
   
   x = ip->x + ip->width - 4;
   y = ip->y + ip->height - 1;
   
   on_icon = ((SLWdg_Mouse_Y == y)
	      && (SLWdg_Mouse_X >= x)
	      && (SLWdg_Mouse_X <= x + 2));
   
   if (on_icon) 
     {
	pb = ip->widget.popbox;
	pb_select (ip, 1);
	return 0;
     }
   return -1;
}


static int popbox_focus_proc (SLWdg_Type *ip, int keysym)
{
   Popup_ListBox_Type *pb;

   switch (keysym)
     {
      case KEY_RETURN:
	pb_select (ip, 1);
	break;
	
      case KEY_MOUSE0:
      case KEY_MOUSE1:
      case KEY_MOUSE2:
	return pb_mouse_select (ip);
	
      case KEY_CTRL_Z:
      case KEY_CTRL_R:
	return -1;
	
      case SLW_DESTROY:
	pb = ip->widget.popbox;
	if (pb != NULL)
	  {
	     if (pb->text != NULL) SLFREE (pb->text);
	     SLFREE (pb);
	     ip->widget.popbox = NULL;
	  }
	break;
	
      case SLW_LOSE_FOCUS:
      default:
	return -1;
     }
   
   return 0;
}

static char *pb_get_text (SLWdg_Type *ip)
{
   Popup_ListBox_Type *pb = ip->widget.popbox;
   if (pb->text == NULL) return "";
   return pb->text;
}


int SLwdg_init_popupbox (SLWdg_Type *ip)
{
   Popup_ListBox_Type *pb;
      
   if (NULL == (pb = (Popup_ListBox_Type *) SLMALLOC (sizeof (List_Box_Type))))
     {
	return -1;
     }
   memset ((char *) pb, 0, sizeof (Popup_ListBox_Type));
   ip->widget.popbox = pb;
   ip->draw = draw_popup_box;
   ip->focus_proc = popbox_focus_proc;
   ip->get_text = pb_get_text;
   ip->set_text = set_popbox_text;
   if (ip->height < 3) ip->height = 3;
   if (ip->width < 3) ip->height = 3;
   pb->lb_id = -1;
   pb->text = NULL;
   if (ip->summary == NULL) ip->summary = "Press Return to make selection.";
   return 0;
}


