#include <stdio.h>
#include <string.h>

#include <stdlib.h>

#include <slang.h>

#include "slwdg.h"
#include "_slwdg.h"

SLWdg_Type *SLWdg_Root_Widget;
/* SLWdg_Type *SLWdg_Focus_Widget; */


#if 0
static void draw_help (SLWdg_Type *ip)
{
   char *msg = ip->help;
   int r, c, dr, dc;
   
   if (msg == NULL) 
     {
	msg = "Help on this item\nis not available\nat this time.";
     }
   
   r = ip->y + 2; c = ip->x + 2;
   
   compute_box_size (msg, &dr, &dc);
   if (SLtt_Use_Ansi_Colors) SLsmg_set_color (FIELD_COLOR);
   SLsmg_write_wrapped_string (msg, r + 1, c + 1, dr, dc, 1);
   
   if (SLtt_Use_Ansi_Colors) 
     {
	if (ip == ip->parent->current)
	  SLsmg_set_color (ACTIVE_COLOR);
	else SLsmg_set_color (BOX_COLOR);
     }
   SLsmg_draw_box (r, c, dr + 2, dc + 2);
   
   if (ip == ip->parent->current) SLsmg_set_color (ACTIVE_COLOR);
   else if (SLtt_Use_Ansi_Colors)
     SLsmg_set_color (LABEL_COLOR);
   
   SLsmg_gotorc (r, c + 1);
   SLsmg_write_string ("[Help]");
   
   SLsmg_gotorc (r + 1, c + 1);
   SLsmg_set_color (0);
}

#endif

static void draw_static_text (SLWdg_Type *ip)
{
   if (ip->label == NULL) return;
   SLsmg_gotorc (ip->y, ip->x);
   SLsmg_write_string (ip->label);
}

static int set_label (SLWdg_Type *ip, char *s)
{
   if (ip->label != NULL) SLFREE (ip->label);
   ip->label = (char *) SLMALLOC (strlen (s) + 1);
   if (ip->label == NULL) return -1;
   strcpy (ip->label, s);
   return 0;
}

int SLwdg_set_label (int id, char *s)
{
   SLWdg_Type *ip;
   if (NULL == (ip = SLwdg_validate_id (id, 0))) return -1;
   return set_label (ip, s);
}

	
/* Note:  In this routine, the current one gets put down last. This is 
 * very important since it means that the current window will be the top one.
 */
void SLwdg_update_display (void)
{
   if (SLwdg_TT_Initialized == 0) return;
   /* The root is a combo box.  Draw it. */
   (*SLWdg_Root_Widget->draw) (SLWdg_Root_Widget);
   SLsmg_refresh ();
}




int SLwdg_focus_next (SLWdg_Type *curr_ip)
{
   SLWdg_Type *ip;
   
   ip = curr_ip;
   do
     {
	ip = ip->next;
	if ((ip->flags & (NO_FOCUS | WIDGET_NOT_MAPPED)) == 0) break;
     }
   while (ip != curr_ip);
   if (ip == curr_ip) return -1;
   ip->parent->current = ip;
   return 0;
}



#define MAX_WIDGETS 128
static SLWdg_Type *Widgets[MAX_WIDGETS];


static int default_resize_widget (SLWdg_Type *ip, int x, int y, int dx, int dy)
{
   SLWdg_Type *parent;
   /* Root is always at 0, 0 */
   if (ip->id != 0)
     {
	parent = ip->parent;
	ip->x = ip->parent->x + x;
	ip->y = ip->parent->y + y;
     }
   ip->width = dx;
   ip->height = dy;
   return 0;
}

/* These are primitive widgets.  Compound ones NOT allowed.
 */
int SLwdg_define_widget (int parent_id,
			 unsigned int type,
			    char *label, 
			    char *text,
			    char *summary,
			    int x, int y, int dx, int dy,
			    int visible)
{
   static int next_id = 0, this_id;
   SLWdg_Type *ip, *parent;
   
   if (next_id == MAX_WIDGETS) 
     {
	/* We have used up all ths slots.  So, brute force to look for 
	 * deleted ones.
	 */
        for (this_id = 0; this_id < MAX_WIDGETS; this_id++)
	  {
	     if (Widgets[this_id] == NULL) break;
	  }
	if (this_id == MAX_WIDGETS) return -1;
     }
   else this_id = next_id++;
   
   ip = (SLWdg_Type *) SLMALLOC (sizeof (SLWdg_Type));
   if (ip == NULL) return -1;
   
   memset ((char *) ip, 0, sizeof (SLWdg_Type));
   
   ip->type = type;
   
   if (label != NULL)
     {
	(void) set_label (ip, label);
     }
   
   if (summary != NULL)
     {
	ip->summary = (char *) SLMALLOC (strlen (summary) + 1);
	if (ip->summary != NULL) strcpy (ip->summary, summary);
     }
   
   ip->x = x;
   ip->y = y;
   ip->flags = 0;
   ip->height = dy;
   ip->width = dx;
   
   if (ip->label != NULL)
     {
	dx = strlen (ip->label) + 4;
     }
   else dx = 0;
   
   if (ip->width < dx) ip->width = dx;
   
   if (visible == 0) ip->flags |= WIDGET_NOT_MAPPED;
   
   if (SLWdg_Root_Widget == NULL)
     {
	SLWdg_Root_Widget = ip;
     }
   
   /* Set up pointer to the parent */
   if (parent_id == 0) 
     {
	ip->parent = SLWdg_Root_Widget;
     }
   else
     {
	if (NULL == (ip->parent = SLwdg_validate_id (parent_id, COMBO_BOX)))
	  {
	     SLFREE (ip);
	     return -1;
	  }
     }
   parent = ip->parent;

   ip->resize_widget = default_resize_widget;
   
   switch (type)
     {
      case COMBO_BOX:
	if (-1 == SLwdg_init_combobox (ip)) goto free_error;
	break;
	
      case EDIT_BOX:
	if (-1 == SLwdg_init_editbox (ip)) goto free_error;
	break;
	
      case LIST_BOX:
	if (-1 == SLwdg_init_listbox (ip)) goto free_error;
	break;
	
      case POPUP_BOX:
	if (-1 == SLwdg_init_popupbox (ip)) goto free_error;
	break;
	
      case STATIC_TEXT:
	ip->draw = draw_static_text;
	ip->set_text = set_label;
	ip->flags |= NO_FOCUS;
	ip->focus_proc = NULL;
	ip->height = 1;
	break;
	
      case BUTTON:
	if (-1 == SLwdg_init_button (ip)) goto free_error;
	break;
	
      default:
	goto free_error;
     }

   Widgets[this_id] = ip;
   
   if (0 == (ip->id = this_id))
     {
	/* This is the root */
	ip->next = ip->prev = ip;
	return 0;
     }
   
   if (parent->child == NULL)
     {
	parent->child = ip;
     }
   
   ip->next = parent->child;
   ip->prev = parent->child->prev;
   ip->next->prev = ip;
   ip->prev->next = ip;
   
   if ((parent->current == NULL)
       && ((ip->flags && NO_FOCUS) == 0)
       && ((ip->flags && WIDGET_NOT_MAPPED) == 0))
     parent->current = ip;
   
   ip->id = this_id;
   
   /* Now fixup x,y coords to absolute ones */
   ip->x += parent->x;
   ip->y += parent->y;
   
   if ((text != NULL) && (ip->set_text != NULL))
     {
	(*ip->set_text) (ip, text);
     }

   return ip->id;
   
   free_error:
   if (ip != NULL)
     {
	if (ip->label != NULL) SLFREE (ip->label);
	if (ip->summary != NULL) SLFREE (ip->summary);
	SLFREE (ip);
     }
   return -1;
}

void SLwdg_delete_widget (SLWdg_Type *ip)
{
   if (ip->num_references) 
     {
	ip->do_free = 1;
	return;
     }
   Widgets[ip->id] = NULL;
   
   ip->num_references = 1;	       /* this is so focus_proc does not 
					* free ip
					*/
   
   if (ip->focus_proc != NULL)
     (*ip->focus_proc) (ip, SLW_DESTROY);
   
   if (ip->label != NULL) SLFREE (ip->label);
   if (ip->summary != NULL) SLFREE (ip->summary);
     
   
   memset ((char *) ip, 0xFF, sizeof (SLWdg_Type));
   SLFREE (ip);
}


int SLwdg_destroy_widget (int id)
{
   SLWdg_Type *ip;
   
   ip = SLwdg_validate_id (id, 0);
   if (ip == NULL) return -1;
   
   if (ip->parent->current == ip)
     {
	/* pop it down to return proper focus */
	SLwdg_popup_widget (id, 0);
     }
   
   /* Now lets figure out what pointers are pointing at us and adjust those. */
   
   ip->next->prev = ip->prev;
   ip->prev->next = ip->next;
     
   if (ip->parent->child == ip)
     {
	if (ip->next == ip)
	  ip->parent->child = NULL;
	else
	  ip->parent->child = ip->next;
     }
   /* No last pointers should be pointing at us. */
   
   SLwdg_delete_widget (ip);
   return 0;
}


SLWdg_Type *SLwdg_validate_id (int id, unsigned int type)
{
   SLWdg_Type *ip;
   
   if ((id < 0) || (id >= MAX_WIDGETS)) 
     return NULL;
   
   ip = Widgets[id];
   if (ip == NULL) return NULL;
   if (type && (ip->type !=  type)) return NULL;
   
   return ip;
}

				  
int SLwdg_set_size (int id, int x, int y, int dx, int dy)
{
   SLWdg_Type *ip;
   if (NULL == (ip = SLwdg_validate_id (id, 0))) 
     return -1;
   
   return (*ip->resize_widget) (ip, x, y, dx, dy);
}

int SLwdg_get_size (int id, int *x, int *y, int *dx, int *dy)
{
   SLWdg_Type *ip, *parent;
   
   if (NULL == (ip = SLwdg_validate_id (id, 0))) 
     return -1;
   
   /* Root is always at 0, 0 */
   if (id != 0)
     {
	parent = ip->parent;
	*x = ip->x - ip->parent->x;
	*y = ip->y - ip->parent->y;
     }
   *dx = ip->width;
   *dy = ip->height;
   return 0;
}


int SLwdg_popup_widget (int id, int state)
{
   SLWdg_Type *ip, *parent, *last;
   
   if (NULL == (ip = SLwdg_validate_id (id, 0))) 
     return -1;
   
   if (ip->flags & NO_FOCUS)
     {
	return -1;
     }
   
   parent = ip->parent;
   if (state == 0)
     {
	if (ip->flags & WIDGET_NOT_MAPPED) return 0;
	ip->flags |= WIDGET_NOT_MAPPED;
	if (ip == parent->current)
	  {
	     last = ip->last;
	     if ((last == NULL)
		 || (last->flags & NO_FOCUS)
		 || (last->flags & WIDGET_NOT_MAPPED))
	       return SLwdg_focus_next (ip);
	     else 
	       {
		  (*ip->focus_proc) (ip, SLW_LOSE_FOCUS);
		  (*last->focus_proc) (last, SLW_GAIN_FOCUS);
		  parent->current = last;
	       }
	  }
     }
   else
     {
	last = ip->last = parent->current;
	if (ip != last) (*last->focus_proc) (last, SLW_LOSE_FOCUS);
	(*ip->focus_proc) (ip, SLW_GAIN_FOCUS);
	parent->current = ip;
	ip->flags &= ~WIDGET_NOT_MAPPED;
     }
   return 0;
}
   
int SLwdg_perform_callback (SLWdg_Type *ip)
{  
   int ret;
   if (ip->callback == NULL) return -1;
   if (ip->flags & PERFORMING_CALLBACK) return -1;
   ip->flags |= PERFORMING_CALLBACK;
   
   ret = (*ip->callback) (ip->id, ip->callback_int, ip->callback_str);
   
   ip->flags &= ~PERFORMING_CALLBACK;
   return ret;
}

int SLwdg_set_focus (int id)
{
   SLWdg_Type *ip;
   
   if (NULL == (ip = SLwdg_validate_id (id, 0))) 
     return -1;
   
   if (ip->flags & (NO_FOCUS | WIDGET_NOT_MAPPED))
     {
	return -1;
     }
   (*ip->parent->current->focus_proc) (ip->parent->current, SLW_LOSE_FOCUS);
   (*ip->focus_proc) (ip, SLW_GAIN_FOCUS);
   ip->parent->current = ip;
   return 0;
}

int SLwdg_set_callback (int id, int (*cb) (int, int, char *), int x, char *str)
{
   SLWdg_Type *ip;
   if (NULL == (ip = SLwdg_validate_id (id, 0))) 
     return -1;
   
   ip->callback = cb;
   ip->callback_int = x;
   ip->callback_str = str;
   return 0;
}


int SLwdg_end (void)
{
   /* SLwdg_popup_widget (0, 0); */
   SLwdg_reset_display ();
   return 0;
}

SLWdg_Type *SLwdg_find_widget_at_xy  (SLWdg_Type *ip_curr, int x, int y)
{
   SLWdg_Type *ip;

   ip = ip_curr;
   
   do
     {
	if (((ip->flags & NO_FOCUS) == 0)
	    && ((ip->flags & WIDGET_NOT_MAPPED) == 0)
	    && (ip->x <= x) && (x < ip->x + ip->width)
	    && (ip->y <= y) && (y < ip->y + ip->height))
	  {
	     return ip;
	  }
	ip = ip->prev;
     }
   while (ip != ip_curr);
   return NULL;
}

void SLwdg_redraw (void)
{
   SLsmg_cls ();
   SLwdg_update_display ();
}

SLKeyMap_List_Type *SLWdg_KeyMap;

int SLWdg_Mouse_X;
int SLWdg_Mouse_Y;

int SLwdg_read_mouse_keys (void)
{
   unsigned char ch;
   int button;
   
   ch = SLang_getkey ();
   switch (ch)
     {
      case ' ':
	button = KEY_MOUSE0;
	break;
	
      case '!':
	button = KEY_MOUSE1;
	break;
	
      case '"':
	button = KEY_MOUSE2;
	break;
	
      default:
	SLang_flush_input ();
	return -1;
     }
   
   SLWdg_Mouse_X = SLang_getkey () - '!';
   SLWdg_Mouse_Y = SLang_getkey () - '!';
   return button;
}

#if 0
static int key_self (void)
{
   if ((SLang_Last_Key_Char == SLang_Abort_Char)
       && SLKeyBoard_Quit)
     {
	SLang_Error = 0;
	SLKeyBoard_Quit = 0;
     }
   return SLang_Last_Key_Char;
}
#endif

char *SLwdg_get_text (int id)
{
   SLWdg_Type *ip;
   
   ip = SLwdg_validate_id (id, 0);
   if ((ip == NULL) || (ip->get_text == NULL)) return NULL;
   return (*ip->get_text) (ip);
}

int SLwdg_set_text (int id, char *s)
{
   SLWdg_Type *ip;
   
   ip = SLwdg_validate_id (id, 0);
   if ((ip == NULL) || (ip->set_text == NULL)) return -1;
   return (*ip->set_text) (ip, s);
}

int SLwdg_widget_state (int id)
{
   SLWdg_Type *ip;
   
   if (NULL == (ip = SLwdg_validate_id (id, 0))) return -1;
   if (ip->flags & WIDGET_NOT_MAPPED) return 0;
   return 1;
}

void SLwdg_take_input (int this_id)
{
   SLWdg_Type *ip, *this_ip;
   SLang_Key_Type *key;
   int keysym;
   int no_update = 0;
   
   if (SLang_Error) return;
   
   if (NULL == (this_ip = SLwdg_validate_id (this_id, COMBO_BOX)))
     return;
   
   if (this_ip->flags & WIDGET_NOT_MAPPED)
     {
	SLwdg_exit_error ("COMBOBOX(id = %d) NOT Mapped!! Cannot take focus.", 
			  this_id);
     }
   
   this_ip->num_references++;
   
   while ((!SLang_Error && (this_ip->flags & WIDGET_NOT_MAPPED) == 0)
	  && SLwdg_TT_Initialized)
     {
	ip = this_ip;
	if (SLang_input_pending (0)) 
	  {
	     no_update = 1;
	  }
	
	if (no_update == 0) SLwdg_update_display ();
	
	if ((NULL == (key = SLang_do_key (SLWdg_KeyMap, (int (*)(void)) SLang_getkey)))
	    || ((key->type != SLKEY_F_KEYSYM) 
		&& (key->f.f == NULL)))
	  {
	     SLtt_beep ();
	     SLang_flush_input ();
	     no_update = 0;
	     continue;
	  }
	
	no_update = 0;
	if (key->type == SLKEY_F_KEYSYM)
	  keysym = key->f.keysym;
	else 
	  keysym = ((int (*)(void)) key->f.f) ();
	
	if (0 == ip->focus_proc (ip, keysym)) continue;
	SLtt_beep ();
	no_update = 1;
     }
   
   this_ip->num_references--;
   if (this_ip->do_free) SLwdg_delete_widget (this_ip);
   SLwdg_update_display ();
}

static int setup_keymaps (void)
{
   int ch;
   char simple[2];
   
   if (SLWdg_KeyMap != NULL) return 0;
   
   if (NULL == (SLWdg_KeyMap = SLang_create_keymap ("_Widget", NULL)))
     return -1;
   
   simple[1] = 0;
   for (ch = 0; ch < 256; ch++)
     {
	simple[0] = (char) ch;
	SLkm_define_keysym (simple, ch, SLWdg_KeyMap);
     }

   
   SLkm_define_keysym ("\033[A", KEY_UP_ARROW, SLWdg_KeyMap);
   SLkm_define_keysym ("\033[B", KEY_DOWN_ARROW, SLWdg_KeyMap);
   SLkm_define_keysym ("\033[C", KEY_RIGHT_ARROW, SLWdg_KeyMap);
   SLkm_define_keysym ("\033[D", KEY_LEFT_ARROW, SLWdg_KeyMap);
   SLkm_define_keysym ("^@", SERVER_ESCAPE, SLWdg_KeyMap);
   SLkm_define_key ("\033[M", (FVOID_STAR) SLwdg_read_mouse_keys, SLWdg_KeyMap);
   
   if (SLang_Error) return -1;

   return 0;
}

int SLwdg_start (void)
{
   SLWdg_Type *ip;
   
   if (-1 == setup_keymaps ()) return -1;
   
   if ((SLWdg_Root_Widget == NULL)
       || (NULL == (ip = SLWdg_Root_Widget->current)))
     return -1;
   
   SLwdg_init_display ();
   /* SLwdg_popup_widget (0, 1); */
   
   if ((ip->flags & NO_FOCUS)
       || (ip->flags & WIDGET_NOT_MAPPED))
     {
	if (-1 == SLwdg_focus_next (ip)) return -1;
     }
   
   /* SLang_Ignore_User_Abort = 1; */
   SLwdg_take_input (0);
   return 0;
}

   
