/* S-Lang Text-Widgets */
/* Copyright (c) 1992, 1995 John E. Davis
 * All rights reserved.
 * 
 * You may distribute under the terms of either the GNU General Public
 * License or the Perl Artistic License.
 */
#include <stdio.h>
#include <slang.h>

#include <stdlib.h>

#include "slwdg.h"


static void initialize (int *color)
{
   SLwdg_initialize (*color);
}

static void init_display (void)
{
   SLwdg_init_display ();
   SLwdg_redraw ();
}

static void reset_display (void)
{
   SLwdg_reset_display ();
}

static void set_focus (int *id)
{
   if (-1 == SLwdg_set_focus (*id))
     {
	SLang_Error = INTRINSIC_ERROR;
     }
}


static void set_size (int *id, int *x, int *y, int *dx, int *dy)
{
   if (-1 == SLwdg_set_size (*id, *x, *y, *dx, *dy)) 
     SLang_Error = INTRINSIC_ERROR;
}

static void get_size (int *id)
{
   int dx, dy, x, y;
   if (-1 == SLwdg_get_size (*id, &x, &y, &dx, &dy))
     SLang_Error = INTRINSIC_ERROR;
   SLang_push_integer (x);
   SLang_push_integer (y);
   SLang_push_integer (dx);
   SLang_push_integer (dy);
}

static int define_widget (void)
{
   int parent, type, x, y, dx, dy, mapped;
   char *label = NULL, *text = NULL, *help = NULL;
   int free_label = 0, free_text = 0, free_help = 0;
   int ret_val;
   
   if (!SLang_pop_integer (&mapped)
       && !SLang_pop_integer (&dy)
       && !SLang_pop_integer (&dx)
       && !SLang_pop_integer (&y)
       && !SLang_pop_integer (&x)
       && !SLang_pop_string (&help, &free_help)
       && !SLang_pop_string (&text, &free_text)
       && !SLang_pop_string (&label, &free_label)
       && !SLang_pop_integer (&type)
       && !SLang_pop_integer (&parent))
     {
	ret_val = SLwdg_define_widget (parent, type,
				       (*label == 0) ? NULL : label,
				       (*text == 0) ? NULL : text,
				       (*help == 0) ? NULL : help,
				       x, y, dx, dy,
				       mapped);
	
     }
   
   if (free_help) SLFREE (help);
   if (free_text) SLFREE (text);
   if (free_label) SLFREE (label);
   
   return ret_val;
}

static void message (char *s)
{
   SLwdg_message (s);
}

static void sl_error_routine (char *s)
{
   int save_error = SLang_Error;
   SLang_Error = 0;
   SLwdg_message (s);
   SLang_Error = save_error;
}
static int Callback_Level;

static int callback_handler (int id, int val, char *str)
{
   SLang_Name_Type *f;
   int ret = -1;
   
   f = (SLang_Name_Type *) str;
   
   SLang_push_integer (id);
   SLang_push_integer (val);
   
   Callback_Level++;
   SLexecute_function (f);
   Callback_Level--;
   
   if (SLang_Error)
     {
	char buf[128];
	sprintf (buf, "Error while executing callback: %s.", f->name);
	SLang_doerror (buf);
	SLwdg_end ();
	Callback_Level--;
	exit (SLang_Error);
     }

   SLang_pop_integer (&ret);
   return ret;
}

static void set_callback (int *id, char *fun, int *val)
{
   SLang_Name_Type *f;
   
   if (NULL == (f = SLang_get_function (fun)))
     {
	SLang_Error = INTRINSIC_ERROR;
	return;
     }
   if (-1 == SLwdg_set_callback (*id, callback_handler, *val, (char *) f))
     SLang_Error = INTRINSIC_ERROR;
}

static void start_slwdg (void)
{
   if (Callback_Level)
     {
	SLang_doerror ("Starting Widgets from a callback is forbidden.");
	return;
     }
   if (-1 == SLwdg_start ()) SLang_Error = INTRINSIC_ERROR;
}

static void end_slwdg (void)
{
   SLwdg_end ();
}

static void set_listbox_items (int *id, char *list)
{
   if (-1 == SLwdg_set_listbox_items (*id, list))
     SLang_Error = INTRINSIC_ERROR;
}

static char *get_text (int *id)
{
   char *txt;
   
   txt = SLwdg_get_text (*id);
   if (txt == NULL) 
     SLang_Error = INTRINSIC_ERROR;
   return txt;
}

static void set_text (int *id, char *s)
{
   if (-1 == SLwdg_set_text (*id, s))
     SLang_Error = INTRINSIC_ERROR;
}

static void set_label (int *id, char *s)
{
   if (-1 == SLwdg_set_label (*id, s))
     SLang_Error = INTRINSIC_ERROR;
}

static void popup_widget (int *id, int *state)
{
   if (-1 == SLwdg_popup_widget (*id, *state))
     {
	return;
     }
}

static void destroy_widget (int *id)
{
   if (-1 == SLwdg_destroy_widget (*id))
     {
	SLang_Error = INTRINSIC_ERROR;
     }
}

static void take_input (int *id)
{
   if (-1 == SLwdg_popup_widget (*id, 1))
     {
	SLang_Error = INTRINSIC_ERROR;
     }
   else SLwdg_take_input (*id);
}

static void lb_set_selection (int *id, char *s)
{
   if (-1 == SLwdg_lb_set_selection (*id, s))
     {
	SLang_Error = INTRINSIC_ERROR;
     }
}

static int widget_state (int *id)
{
   return SLwdg_widget_state (*id);
}


static int ButtonType = BUTTON;
static int ComboBoxType = COMBO_BOX;
static int EditBoxType = EDIT_BOX;
static int ListBoxType = LIST_BOX;
static int StaticTextType = STATIC_TEXT;
static int PopupListBoxType = POPUP_BOX;

#define I SLANG_INT_TYPE
static SLang_Intrin_Fun_Type SLwdg_Intrinsics [] =
{
   MAKE_INTRINSIC_I("slwdg_initialize", initialize, VOID_TYPE),
   MAKE_INTRINSIC_0("slwdg_start", start_slwdg, VOID_TYPE),
   MAKE_INTRINSIC_0("slwdg_end", end_slwdg, VOID_TYPE),
   MAKE_INTRINSIC_0("slwdg_reset_display", reset_display, VOID_TYPE),
   MAKE_INTRINSIC_0("slwdg_init_display", init_display, VOID_TYPE),
   MAKE_INTRINSIC_I("slwdg_set_focus", set_focus, VOID_TYPE),
   MAKE_INTRINSIC_S("slwdg_message", message, VOID_TYPE),
   MAKE_INTRINSIC_0("slwdg_define_widget", define_widget, INT_TYPE),
   /* Prototype: Integer slwdg_define_widget (Integer parent,
    * @ 				      Integer type,
    * @ 				      String label,
    * @ 				      String help,
    * @ 				      Integer x, y, dx, dy,
    * @ 				      Integer mapped);
    * This function is used to define a widget.  The nine parameters have
    * the following meaning:
    * @ parent        The parent id  (A COMBOBOX)
    * @ type          Type of widget that is being defined
    * @ label         Text that is associated with the widget
    * @ help          A help string that describes the widget
    * @ x, y          Coordinates of the upper left corner of the widget
    * @ dx, dy        Width and height of the widget
    * @ mapped        If non-zero, the widget will be visible when the parent
    * @                 becomes visible
    * Most of the fields are self explanatory.  However, two require further 
    * explanation.  The @type@ parameter must be one of the following values:
    * @ COMBO_BOX
    * @ LIST_BOX
    * @ EDIT_BOX
    * @ BUTTON
    * @ STATIC_TEXT
    * @ POPUP_BOX
    * The @label@ parameter has a different meaning that varies with the
    * type of widget.  For example if the widget type is @BUTTON@, then
    * @label@ specifies the name of the button.
    *
    * The @parent@ parameter must be the id of a @COMBO_BOX@.  The root widget
    * is a predefined @COMBOX_BOX@ with an id of 0.
    */
   MAKE_INTRINSIC_ISI("slwdg_set_callback", set_callback, VOID_TYPE),
   MAKE_INTRINSIC_S("slwdg_message", SLwdg_message, VOID_TYPE),
   MAKE_INTRINSIC_I("slwdg_get_text", get_text, STRING_TYPE),
   MAKE_INTRINSIC_IS("slwdg_set_text", set_text, VOID_TYPE),
   MAKE_INTRINSIC_IS("slwdg_set_label", set_label, VOID_TYPE),
   MAKE_INTRINSIC_S("slwdg_get_yes_no_cancel", SLwdg_get_yes_no_cancel, INT_TYPE),
   MAKE_INTRINSIC_IS("slwdg_set_listbox_items", set_listbox_items, VOID_TYPE),
   MAKE_INTRINSIC_IS("slwdg_set_listbox_selection", lb_set_selection, VOID_TYPE),
   MAKE_INTRINSIC_II("slwdg_popup_widget", popup_widget, VOID_TYPE),
   MAKE_INTRINSIC_I("slwdg_take_input", take_input, VOID_TYPE),
   MAKE_INTRINSIC_I("slwdg_get_size", get_size, VOID_TYPE),
   MAKE_INTRINSIC_5("slwdg_set_size", set_size, VOID_TYPE, I,I,I,I,I),
   MAKE_INTRINSIC_I("slwdg_destroy_widget", destroy_widget, VOID_TYPE),
   MAKE_INTRINSIC_I("slwdg_widget_state", widget_state, INT_TYPE),
   
   SLANG_END_TABLE
};

static SLang_Intrin_Var_Type SLwdg_Variables [] = 
{
   MAKE_VARIABLE ("COMBO_BOX", &ComboBoxType, INT_TYPE, 1),
   MAKE_VARIABLE ("BUTTON", &ButtonType, INT_TYPE, 1),
   MAKE_VARIABLE ("EDIT_BOX", &EditBoxType, INT_TYPE, 1),
   MAKE_VARIABLE ("LIST_BOX", &ListBoxType, INT_TYPE, 1),
   MAKE_VARIABLE ("STATIC_TEXT", &StaticTextType, INT_TYPE, 1),
   MAKE_VARIABLE ("POPUP_BOX", &PopupListBoxType, INT_TYPE, 1),
   MAKE_VARIABLE ("SCREEN_ROWS", &SLtt_Screen_Rows, INT_TYPE, 1),
   MAKE_VARIABLE ("SCREEN_COLS", &SLtt_Screen_Cols, INT_TYPE, 1),
   SLANG_END_TABLE
};
#undef I
int init_SLwdg (void)
{
   SLang_Error_Hook = sl_error_routine;
 
   if ((-1 == SLadd_intrin_fun_table (SLwdg_Intrinsics, "SLWDG"))
       || (-1 == SLadd_intrin_var_table (SLwdg_Variables, NULL)))
     return -1;
   
   return 0;
}
