#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <stdarg.h>

#ifndef NO_UNISTD_H
#include <unistd.h>
#endif

#ifndef sun
#include <sys/ioctl.h>
#endif

#ifdef HAS_TERMIOS
# include <termios.h>
#endif
#ifdef SYSV
# include <sys/termio.h>
# include <sys/stream.h>
# include <sys/ptem.h>
# include <sys/tty.h>
#endif

#include "slang.h"
#include "slwdg.h"
#include "_slwdg.h"


static void widget_hangup (int sig)
{
   exit (sig);
}


int SLwdg_TT_Initialized = 0;

void SLwdg_suspend (int sig)
{
   (void) sig;
   SLwdg_reset_display ();
   (void) kill(getpid(),SIGTSTP);
   SLwdg_init_display ();
   SLwdg_redraw ();
}

void (*Forms_Winch_Function)(void);

static void widget_set_screen_size (int sig)
{
   int r = 0, c = 0;
   
#ifdef TIOCGWINSZ
   struct winsize wind_struct;
   
   
   if ((ioctl(1,TIOCGWINSZ,&wind_struct) == 0)
       || (ioctl(0, TIOCGWINSZ, &wind_struct) == 0)
       || (ioctl(2, TIOCGWINSZ, &wind_struct) == 0))
     {
	c = (int) wind_struct.ws_col;
	r = (int) wind_struct.ws_row;
     }

#endif
   
   if ((r <= 0) || (r > 200)) r = 24;
   if ((c <= 0) || (c > 250)) c = 80;
   SLtt_Screen_Rows = r;
   SLtt_Screen_Cols = c;
   
#ifdef SIGWINCH
   if (sig == SIGWINCH) 
     {
	SLsmg_reset_smg ();
	SLsmg_init_smg ();
	if (Forms_Winch_Function != NULL) (*Forms_Winch_Function) ();
	SLwdg_redraw ();
     }
   signal (SIGWINCH, widget_set_screen_size);
#endif
}

static void abort_fun (int sig)
{
   (void) sig;
   signal (SIGINT, abort_fun);
   if (SLang_Ignore_User_Abort == 0)
     {
	SLang_Error = USER_BREAK;
	SLwdg_exit_error ("Abort key pressed.");
     }
}


int SLwdg_init_display (void)
{
   if (SLwdg_TT_Initialized) return -1;
   SLwdg_TT_Initialized = 1;
   SLang_init_tty (7, 1, 0);
   SLang_set_abort_signal (abort_fun);
   widget_set_screen_size (0);
   SLsmg_init_smg ();
   
   SLtt_set_mouse_mode (1, 0);
   
   signal(SIGTSTP, SLwdg_suspend);
#ifdef SIGPIPE
   signal(SIGPIPE, SIG_IGN);
#endif
#ifdef SIGTTOU
   /* Allow background writes */
   signal (SIGTTOU, SIG_IGN);
#endif
   signal (SIGHUP, widget_hangup);
   return 0;
}

int SLwdg_reset_display ()
{
   if (SLwdg_TT_Initialized == 0) return -1;
   SLwdg_TT_Initialized = 0;
   SLtt_set_mouse_mode (0, 0);
   SLsmg_gotorc (SLtt_Screen_Rows - 1, 0);
   SLsmg_refresh ();
   SLsmg_reset_smg ();
   SLang_reset_tty ();
   signal(SIGTSTP, SIG_DFL);
   return 0;
}


int SLwdg_initialize (int color)
{
   char *bg = "black";
   
   SLtt_get_terminfo();
   widget_set_screen_size (0);
   if (color)
     {
	SLtt_Use_Ansi_Colors = color;
     }
   SLtt_set_color (0, NULL, "lightgray", bg);
   SLtt_set_color (LABEL_COLOR, NULL, "cyan", bg);
   SLtt_set_color (ACTIVE_LABEL_COLOR, NULL, "yellow", bg);
   SLtt_set_mono (LABEL_COLOR, NULL, 0);
   SLtt_set_mono (ACTIVE_LABEL_COLOR, NULL, SLTT_REV_MASK);
   SLtt_set_color (FIELD_COLOR, NULL, "lightgray", bg);
   SLtt_set_mono (FIELD_COLOR, NULL, 0);
   SLtt_set_color (BOX_COLOR, NULL, "green", bg);
   SLtt_set_color (ACTIVE_BOX_COLOR, NULL, "yellow", bg);
   SLtt_set_mono (BOX_COLOR, NULL, 0);
   SLtt_set_mono (ACTIVE_BOX_COLOR, NULL, SLTT_BOLD_MASK);
   SLtt_set_color (ACTIVE_COLOR, NULL, "yellow", bg);
   SLtt_set_mono (ACTIVE_COLOR, NULL, SLTT_BOLD_MASK);
   SLtt_set_color (SUMMARY_COLOR, NULL, "yellow", "blue");
   SLtt_set_mono (SUMMARY_COLOR, NULL, SLTT_REV_MASK);
   /* SLtt_add_color_attribute (SUMMARY_COLOR, SLTT_BLINK_MASK); */
   SLtt_set_color (SELECTION_COLOR, NULL, "yellow", "blue");
   SLtt_set_mono (SELECTION_COLOR, NULL, SLTT_BOLD_MASK);
   SLtt_set_color (BUTTON_COLOR, NULL, "lightgray", "magenta");
   SLtt_set_mono (BUTTON_COLOR, NULL, SLTT_REV_MASK);
   SLtt_set_color (ACTIVE_BUTTON_COLOR, NULL, "yellow", "magenta");
   SLtt_set_mono (ACTIVE_BUTTON_COLOR, NULL, SLTT_REV_MASK | SLTT_BOLD_MASK);
   SLtt_set_color (ACTIVE_FIELD_COLOR, NULL, "white", bg);
   SLtt_set_mono (ACTIVE_FIELD_COLOR, NULL, SLTT_REV_MASK);
   
   return SLwdg_define_widget (0, COMBO_BOX, NULL, NULL, NULL,
				  0, 0, SLtt_Screen_Cols, SLtt_Screen_Rows,
				  1);
}

void SLwdg_exit_error (char *fmt, ...)
{
   char buf[256];
   va_list ap;
   
   SLwdg_reset_display ();
   va_start(ap, fmt);
   vsprintf(buf, fmt, ap);
   SLang_doerror (buf);
   exit (-1);
}

	
